UNIT GenMoves;

INTERFACE
USES
  Globals;


{---------------------------Generate Move list-----------------------------}
CONST
  QuietMoveValue = 190;
TYPE
  MoveListType = ARRAY[0..MaxPly, 0..MaxPlyMoves] OF MoveType;  {-- the move list per ply }
VAR
  MoveList      : MoveListType;                 {-- the move list }
  AllMovesQuiet : Boolean;                      {-- unless proven otherwise }

FUNCTION  GenCaptureMoves : Byte;               {-- gen all capture moves }
FUNCTION  GenAllMoves(TTmove : Str4) : Byte;    {-- as it says. 0 means no moves available. }
FUNCTION  GetAMove(MaxMoves : Byte)  : Str4;    {-- returen the next best move }


{-----------------------------------------------------------------------------}
IMPLEMENTATION
USES
  Utils,
  LookUp,
  Debugging;


{$ASSERTIONS ON}
{$RANGECHECKS ON}

CONST
  CaptureFlag = 190;               {-- flags a capture move }

VAR
  MoveListIndex : Byte;            {-- incremented for each new move }
  HashMove      : Str4;            {-- to be searched first }


PROCEDURE AddMove(FromSq, ToSq : SquareType; Value : Int16);
{-- MoveListIndex is incremented first, so for no moves left, it will be zero
    Adds the move to the move list and gives it a value that is used for picking the next best move
    order should be: Hash
                     Promotion
                     Captures
                     Quiet
}
VAR
  Mstr      : Str4;
  MovPiece,
  CapPiece  : PieceType;
  Promotion : Boolean;
BEGIN
  Assert(MoveListIndex < MaxPlyMoves,   'error proc AddMove, more than 255 moves');
  Assert(OnBoard(FromSq),               'error proc AddMove FromSq');
  Assert(OnBoard(ToSq),                 'error proc AddMove ToSq');

  Inc(MoveListIndex);

  {-- convert to move string }
  Mstr := BoardSquareToString(FromSq) + BoardSquareToString(ToSq);
  {-- add the move to the move list }
  MoveList[Ply, MoveListIndex].MoveStr := Mstr;

  CapPiece  := Data[Ply].Board[ToSq];
  MovPiece  := Data[Ply].Board[FromSq];
  Promotion := ( (MovPiece = WPawn) AND (ToSq > $69) ) OR ( (MovPiece = BPawn) AND (ToSq < 8) );

  {-- now set a rating/score }
  IF Mstr = HashMove THEN  {-- in a PV node this should be the hash move }
    MoveList[Ply, MoveListIndex].MoveValue := 5000
  ELSE
    IF Promotion THEN
      MoveList[Ply, MoveListIndex].MoveValue := 2500
    ELSE
      IF Value = CaptureFlag THEN  {-- some tricks to get MVV/LVA move ordering }
        BEGIN
          {-- first change black pieces to white pieces }
          IF Ord(CapPiece) > 6 THEN
            CapPiece := PieceType(Ord(CapPiece) - 6);
          IF Ord(MovPiece) > 6 THEN
            MovPiece := PieceType(Ord(MovPiece) - 6);
          {-- then the formula...values should run from 199 -> 999 }
          MoveList[Ply, MoveListIndex].MoveValue := 200*Ord(CapPiece) - Ord(MovPiece);
        END
      ELSE     (* killer moves are probably contained in the history scores....so dont do it twice *)
(*        IF Mstr = KillerMoves[0, Ply].MoveStr THEN
          MoveList[Ply, MoveListIndex].MoveValue := 109
       ELSE
         IF Mstr = KillerMoves[1, Ply].MoveStr THEN
           MoveList[Ply, MoveListIndex].MoveValue := 108
         ELSE
*)           {-- the quiet moves have their value already set by PST, so fold in the history }
           MoveList[Ply, MoveListIndex].MoveValue := Value + HistoryTable[MovPiece, ToSq];   {-- are killers in the history table ??}

  {-- test and set all quiet moves flag.....used in LMR }
  IF AllMovesQuiet THEN
    AllMovesQuiet := MoveList[Ply, MoveListIndex].MoveValue < QuietMoveValue;
END;


PROCEDURE GenPawnMoves(Plist : PawnListType);
{-- but not pawn captures. Uses PST for ordering }
VAR
  FromSq,
  ToSq   : SquareType;
  i      : 0..7;
BEGIN
  IF Wtm THEN
    FOR i := 0 TO 7 DO                     {-- cycle through the pawns list }
      BEGIN
        IF Plist[i] <> OffBoard THEN
          BEGIN
            FromSq := Plist[i];            {-- get a pawn }
            IF Rank(FromSq) < 6 THEN       {-- promotion is done in a seperate procedure }
              BEGIN
               {-- one square ahead }
               ToSq := FromSq + $10;
               IF OnBoard(ToSq) AND (Data[Ply].Board[ToSq] = NoPiece) THEN
                 AddMove(FromSq, ToSq, PawnPST[ToSq] - PawnPST[FromSq]);

               {-- two squares ahead }
               IF (FromSq < $18) THEN  {-- on second rank }
                 BEGIN
                   ToSq := FromSq + $20;
                   IF (Data[Ply].Board[FromSq + $10] = NoPiece) AND  {-- nothing directly infront }
                      (Data[Ply].Board[ToSq] = NoPiece) THEN
                     AddMove(FromSq, ToSq, PawnPST[ToSq] - PawnPST[FromSq]);
                 END;
              END;
         END;
      END
  ELSE  {-- blacks move }
    FOR i := 0 TO 7 DO
      BEGIN
        IF Plist[i] <> OffBoard THEN
          BEGIN
            FromSq := Plist[i];
            IF Rank(FromSq) > 1 THEN     {-- promotion is done sepertely }
              BEGIN
                {-- one square ahead }
                ToSq := FromSq - $10;
                IF OnBoard(ToSq) AND (Data[Ply].Board[ToSq] = NoPiece) THEN
                  AddMove(FromSq, ToSq, PawnPST[ToSq+8] - PawnPST[FromSq+8]);

                {-- two squares ahead }
                IF (FromSq > $59) THEN
                  BEGIN
                    ToSq := FromSq - $20;
                    IF (Data[Ply].Board[FromSq - $10] = NoPiece) AND
                       (Data[Ply].Board[ToSq] = NoPiece) THEN
                      AddMove(FromSq, ToSq, PawnPST[ToSq+8] - PawnPST[FromSq+8]);
                  END;
              END;
         END;
      END;
END;


PROCEDURE GenSliderMoves(Plist  : PieceListType; {-- list of squares the pieces are on }
                         OffSet : DirType;       {-- array of the directions a piece can move }
                         d      : Int8);         {-- the number of directions a piece can move }

VAR
  i       : 0..7;
  j       : 0..1;
  FromSq,
  ToSq    : Int16;

BEGIN
  IF (Plist[0] <> OffBoard) OR (Plist[1] <> OffBoard) THEN
    FOR j := 0 TO 1 DO                           {-- each piece }
      BEGIN
        FromSq := Plist[j];                      {-- the pieces origin }
        IF FromSq <> OffBoard THEN
          FOR i := 0 TO d DO                     {-- each direction it can move in }
            BEGIN
              ToSq := FromSq + OffSet[i];        {-- 1st destination square }
              WHILE OnBoard(ToSq) AND (Data[Ply].Board[ToSq] = NoPiece) DO {-- empty squares }
                BEGIN
                  AddMove(FromSq, ToSq, SliderPST[ToSq] - SliderPST[FromSq]);
                  ToSq := ToSq + OffSet[i];      {-- next square }
                END;
            END;
      END;
END;


PROCEDURE GenSliderCaps(Plist  : PieceListType; {-- list of squares the pieces are on }
                        OffSet : DirType;       {-- array of the directions a piece can move }
                        d      : Int8);         {-- the number of directions a piece can move }

VAR
  i       : 0..7;
  j       : 0..1;
  FromSq,
  ToSq    : Int16;
  Pset    : SetOfPieces;
BEGIN
  IF Wtm THEN                                {-- the pieces that white could capture }
    Pset := BlackPieceSet
  ELSE
    Pset := WhitePieceSet;                   {-- the pieces black could capture }

  IF (Plist[0] <> OffBoard) OR (Plist[1] <> OffBoard) THEN
    FOR j := 0 TO 1 DO                         {-- each piece }
      BEGIN
        FromSq := Plist[j];                    {-- the pieces origin }
        IF FromSq <> OffBoard THEN
          FOR i := 0 TO d DO                   {-- each direction it can move in }
            BEGIN
              ToSq := FromSq + OffSet[i];      {-- 1 st destination square }
              WHILE OnBoard(ToSq) AND (Data[Ply].Board[ToSq] = NoPiece) DO  {-- scan for a piece }
                ToSq := ToSq + OffSet[i];

              IF OnBoard(ToSq) AND (Data[Ply].Board[ToSq] IN Pset) THEN
                AddMove(FromSq, ToSq, CaptureFlag);
            END;
      END;
END;


PROCEDURE GenKnightMoves(Plist : PieceListType);
VAR
  d      : 0..7;
  i      : 0..1;
  FromSq,
  ToSq   : Int16;
BEGIN
  IF (Plist[0] <> OffBoard) OR (Plist[1] <> OffBoard) THEN
    FOR i := 0 to 1 DO               {-- knight piece list }
      BEGIN
        FromSq := Plist[i];          {-- origin square }
        IF FromSq <> OffBoard THEN
          FOR d := 0 TO 7 DO         {-- each direction }
            BEGIN
              ToSq := FromSq + KnightOffSet[d];  {-- destination square }
              IF OnBoard(ToSq) AND (Data[Ply].Board[ToSq] = NoPiece) THEN
                AddMove(FromSq, ToSq, KnightPst[ToSq] - KnightPST[FromSq]);
            END;
      END;
END;


PROCEDURE GenKnightCaps(Plist : PieceListType);
VAR
  d      : 0..7;
  i      : 0..1;
  FromSq,
  ToSq   : Int16;
  Pset   : SetOfPieces;
BEGIN
  IF Wtm THEN
    Pset := BlackPieceSet                       {-- enermy pieces }
  ELSE
    Pset := WhitePieceSet;

  IF (Plist[0] <> OffBoard) OR (Plist[1] <> OffBoard) THEN
    FOR i := 0 to 1 DO                          {-- knight piece list }
      BEGIN
        FromSq := Plist[i];                     {-- origin square }
        IF FromSq <> OffBoard THEN
          FOR d := 0 TO 7 DO                    {-- each direction }
            BEGIN
              ToSq := FromSq + KnightOffSet[d];  {-- destination square }
              IF OnBoard(ToSq) AND (Data[Ply].Board[ToSq] IN Pset) THEN  {-- found enemy piece }
                AddMove(FromSq, ToSq, CaptureFlag);
            END;
      END;
END;


FUNCTION CastlingOK(ToSq : SquareType) : Boolean;
{-- checks to see if castling is legal:
    Havent moved king or rook,
    Arent in check,
    No interveining pieces,
    No inbetween squares attacked }
VAR
  OK : Boolean;
BEGIN
  WITH Data[Ply] DO
    BEGIN
      IF (Wtm AND (W_Rook[0] = OffBoard)) OR
         ((NOT Wtm) AND (B_Rook[0] = OffBoard)) THEN
        Exit;

      OK := True;   {-- unless proven otherwise }

      IF OK THEN
        CASE ToSq OF
          6 : OK := (Board[5]   = NoPiece) AND    {-- white castles short }
                    (Board[6]   = NoPiece) AND
                    (Board[7]   = WRook);
          2 : OK := (Board[0]   = WRook)   AND    {-- white castles long }
                    (Board[1]   = NoPiece) AND
                    (Board[2]   = NoPiece) AND
                    (Board[3]   = NoPiece);
        $76 : OK := (Board[$75] = NoPiece) AND    {-- black castles short }
                    (Board[$76] = NoPiece) AND
                    (Board[$77] = BRook);
        $72 : OK := (Board[$70] = BRook)   AND    {-- black castles long }
                    (Board[$71] = NoPiece) AND
                    (Board[$72] = NoPiece) AND
                    (Board[$73] = NoPiece);
        END; {-- case}

      {-- you cant castle if you are in check or if there are inbetween squares attacked }
      IF OK THEN
        IF Wtm THEN
          BEGIN
            IF ToSq = 6 THEN  {-- castling short }
              OK := NOT Attacked(4, BlackPieces) AND  {-- cant castle out of check }
                    NOT Attacked(5, BlackPieces) AND  {-- or across check }
                    NOT Attacked(6, BlackPieces)
            ELSE  {-- castling long }
              OK := NOT Attacked(4, BlackPieces) AND
                    NOT Attacked(2, BlackPieces) AND
                    NOT Attacked(3, BlackPieces)
          END
        ELSE
          BEGIN  {-- blacks move }
            IF ToSq = $76 THEN  {-- castling short }
              OK := NOT Attacked($74, WhitePieces) AND
                    NOT Attacked($75, WhitePieces) AND
                    NOT Attacked($76, WhitePieces)
            ELSE
              OK := NOT Attacked($74, WhitePieces) AND
                    NOT Attacked($72, WhitePieces) AND
                    NOT Attacked($73, WhitePieces);
          END;
    END; {-- with }

  CastlingOK := OK;
END;


PROCEDURE GenCastling;
BEGIN
  IF Wtm THEN
    BEGIN
      IF Data[Ply].W_King = 4 THEN
        BEGIN
          IF Data[Ply].WcanCastShort THEN
            IF CastlingOK(6) THEN
              AddMove(4, 6, 200);

          IF Data[Ply].WcanCastLong THEN
            IF CastlingOK(2) THEN
              AddMove(4, 2, 180);
        END;
    END
  ELSE  {-- blacks move }
    IF (Data[Ply].B_King = $74) THEN
      BEGIN
        IF Data[Ply].BcanCastShort THEN
          IF CastlingOK($76) THEN
            AddMove($74, $76, 200);

        IF Data[Ply].BcanCastLong THEN
          IF CastlingOK($72) THEN
            AddMove($74, $72, 180);
      END;
END;


PROCEDURE GenKingMoves(CapGen : Boolean);
{-- legality is checked in search }
VAR
  d      : 0..7;
  FromSq,
  ToSq   : Int16;
BEGIN
  IF Wtm THEN
    BEGIN
      FromSq := Data[Ply].W_King;             {-- origin square }
      FOR d := 0 TO 7 DO                      {-- each direction }
        BEGIN
          ToSq := FromSq + KingOffset[d];     {-- destination square }
          IF OnBoard(ToSq) THEN
            BEGIN
              IF Not CapGen THEN
                BEGIN
                  IF (Data[Ply].Board[ToSq] = NoPiece) THEN
                    IF EndGame THEN
                      AddMove(FromSq, ToSq, KingPst1[ToSq] - KingPST1[FromSq])
                    ELSE
                      AddMove(FromSq, ToSq, KingPst2[ToSq] - KingPST2[FromSq])
                END
              ELSE  {-- yes capture generation }
                IF (Data[Ply].Board[ToSq] IN BlackPieceSet) THEN
                  AddMove(FromSq, ToSq, CaptureFlag);
            END;
        END;

      IF NOT CapGen THEN
        GenCastling;
    END
  ELSE
    BEGIN
      FromSq := Data[Ply].B_King;             {-- origin square }
      FOR d := 0 TO 7 DO                      {-- each direction }
        BEGIN
          ToSq := FromSq + KingOffset[d];     {-- destination square }
          IF OnBoard(ToSq) THEN
            BEGIN
              IF NOT CapGen THEN
                BEGIN
                  IF (Data[Ply].Board[ToSq] = NoPiece) THEN
                    IF NOT EndGame THEN
                      AddMove(FromSq, ToSq, KingPst1[ToSq] - KingPST1[FromSq])
                    ELSE
                      AddMove(FromSq, ToSq, KingPst2[ToSq] - KingPST2[FromSq])
                END
              ELSE  {-- yes capture generation }
                IF (Data[Ply].Board[ToSq] IN WhitePieceSet) THEN
                  AddMove(FromSq, ToSq, CaptureFlag);
            END;
        END;

      IF NOT CapGen THEN
        GenCastling;
    END;
END;


PROCEDURE GenPawnCaptures;
VAR
  ToSq,
  FromSq : -1..$78;
  i : 0..7;
BEGIN
  IF Wtm THEN
    BEGIN
      FOR i := 0 TO 7 DO
        BEGIN
          FromSq := Data[Ply].W_Pawns[i];  {-- origin square }
          IF FromSq <> OffBoard THEN
            BEGIN
              ToSq := FromSq + $F;             {-- capture left }
              IF OnBoard(ToSq) AND(Data[Ply].Board[ToSq] IN BlackPieceSet) THEN
                AddMove(FromSq, ToSq, CaptureFlag);

              ToSq := FromSq + $11;            {-- capture right }
              IF OnBoard(ToSq) AND(Data[Ply].Board[ToSq] IN BlackPieceSet) THEN
                AddMove(FromSq, ToSq, CaptureFlag);

              IF Data[Ply].EnpasauntSq <> OffBoard THEN  {-- the previous move was a 2 square pawn move }
                IF ((FromSq - 1) = Data[Ply].EnpasauntSq) OR ((FromSq + 1) = Data[Ply].EnpasauntSq) THEN
                  AddMove(FromSq, Data[Ply].EnpasauntSq + $10, CaptureFlag);
            END;
        END;
    END
  ELSE
    BEGIN
      FOR i := 0 TO 7 DO
        BEGIN
          FromSq := Data[Ply].B_Pawns[i];
          IF FromSq <> OffBoard THEN
            BEGIN
              ToSq := FromSq - $F;
              IF OnBoard(ToSq) AND(Data[Ply].Board[ToSq] IN WhitePieceSet) THEN
                AddMove(FromSq, ToSq, CaptureFlag);

              ToSq := FromSq - $11;
              IF OnBoard(ToSq) AND(Data[Ply].Board[ToSq] IN WhitePieceSet) THEN
                AddMove(FromSq, ToSq, CaptureFlag);

              IF Data[Ply].EnpasauntSq <> OffBoard THEN
                IF ((FromSq - 1) = Data[Ply].EnpasauntSq) OR ((FromSq + 1) = Data[Ply].EnpasauntSq) THEN
                  AddMove(FromSq, Data[Ply].EnpasauntSq - $10, CaptureFlag);
            END;
        END;
    END;
END;


PROCEDURE GenPawnPromotion;
CONST
  PawnPromVal = 4000;
VAR
  I : 0..7;
BEGIN
  WITH Data[Ply] DO
    IF Wtm THEN
      BEGIN
        FOR i := 0 TO 7 DO
          IF OnBoard(W_Pawns[i]) AND (Rank(W_Pawns[i]) > 5) THEN
            IF Data[Ply].Board[ W_Pawns[i] + $10 ] = NoPiece THEN
              AddMove(W_Pawns[i], W_Pawns[i] + $10, PawnPromVal);
      END
    ELSE
      BEGIN
        FOR i := 0 TO 7 DO
          IF OnBoard(B_Pawns[i]) AND (Rank(B_Pawns[i]) < 2) THEN
            IF Data[Ply].Board[ B_Pawns[i] - $10 ] = NoPiece THEN
            AddMove(B_Pawns[i], B_Pawns[i] - $10, PawnPromVal);
      END;
END;


FUNCTION GenCaptureMoves : Byte;
{-- used when generating all moves, but also for quiescent search, so pawn promotion
    and king moves are generated here }
BEGIN
  {-- zero this plys list }
  FillChar(MoveList[Ply], SizeOf(MoveList[Ply]), 0);
  MoveListIndex := 0;

  GenPawnCaptures;
  GenPawnPromotion;
  GenKingMoves(True);   {-- king moves that capture }

  IF Wtm THEN
    BEGIN
      GenKnightCaps(Data[Ply].W_Knight);
      GenSliderCaps(Data[Ply].W_Rook,   RookOffset,   3);
      GenSliderCaps(Data[Ply].W_Bishop, BishopOffset, 3);
      GenSliderCaps(Data[Ply].W_Queen,  QueenOffset,  7);
    END
  ELSE
    BEGIN
      GenKnightCaps(Data[Ply].B_Knight);
      GenSliderCaps(Data[Ply].B_Rook,   RookOffset,   3);
      GenSliderCaps(Data[Ply].B_Bishop, BishopOffset, 3);
      GenSliderCaps(Data[Ply].B_Queen,  QueenOffset,  7);
    END;

  GenCaptureMoves := MoveListIndex;
END;


PROCEDURE Gen_Moves(Piece : PieceType);
{-- generates non-capture moves for each piece }
BEGIN
  CASE Piece OF
    Wpawn   : GenPawnMoves  (Data[Ply].W_Pawns);
    Bpawn   : GenPawnMoves  (Data[Ply].B_Pawns);
    Wrook   : GenSliderMoves(Data[Ply].W_Rook,   RookOffset,   3);
    Brook   : GenSliderMoves(Data[Ply].B_Rook,   RookOffset,   3);
    Wbishop : GenSliderMoves(Data[Ply].W_Bishop, BishopOffset, 3);
    Bbishop : GenSliderMoves(Data[Ply].B_Bishop, BishopOffset, 3);
    Wqueen  : GenSliderMoves(Data[Ply].W_Queen,  QueenOffset,  7);
    Bqueen  : GenSliderMoves(Data[Ply].B_Queen,  QueenOffset,  7);
    Wknight : GenKnightMoves(Data[Ply].W_Knight);
    Bknight : GenKnightMoves(Data[Ply].B_Knight);
    Wking,
    Bking   : GenKingMoves(False);
  END; {-- case }
END;


PROCEDURE GenNonCapMoves;
VAR
  p : PieceType;
BEGIN
  IF Wtm THEN
    FOR p := Wpawn TO Wking DO
      Gen_Moves(p)
  ELSE
    FOR p := Bpawn TO Bking DO
      Gen_Moves(p);
END;


FUNCTION GenAllMoves(TTmove : Str4) : Byte;
VAR
  i : Byte;
BEGIN
  AllMovesQuiet := True;   {-- unless proven otherwise }

  IF TTmove <> '' THEN
    HashMove := TTmove
  ELSE
    HashMove := PV[Ply, Ply];

  i := GenCaptureMoves;
  GenNonCapMoves;

  GenAllMoves := MoveListIndex;

{  if ply = 10 then begin writemovelist(10);  halt; end;}
END;


FUNCTION GetAMove(MaxMoves : Byte) : Str4;
{-- scan the move list and pick the next highest score, then flag it as used }
CONST
  Sentinal = -32500;
VAR
  i, j,
  MaxScore : Int16;
BEGIN
  MaxScore := Sentinal;
  j := 1;
  FOR i := 1 TO MaxMoves DO
    IF MoveList[Ply, i].MoveValue > MaxScore THEN
      BEGIN
        MaxScore := MoveList[Ply, i].MoveValue;  {-- note highest score }
        j := i;                                  {-- and save position in the array }
      END;

  MoveList[Ply, j].MoveValue := Sentinal;        {-- flag to say move is used }
  GetAMove := MoveList[Ply, j].MoveStr;          {-- return highest scoring move }
END;


BEGIN
END.
