 UNIT Globals;

INTERFACE

CONST
  MaxPly       =   50;  {-- max search depth }
  MaxPlyMoves  =  100;  {-- max moves at any position }
  MaxGameMoves =  512;  {-- max half moves in the game }
  MaxSquares   =  $7F;  {-- last square is H8 }
  OffBoard     =  $4B;  {-- sentinel value to flag off board = 75d }
  MaxMaterial  = 4250;  {-- assuming 1 queen }

TYPE
  Int8   = ShortInt;    {-- 1 byte  }
  Int16  = SmallInt;    {-- 2 bytes }
  Int32  = LongInt;     {-- 4 bytes }
  Word16 = Word;        {-- 2 bytes }
  Word32 = Cardinal;    {-- 4 bytes }
  Word64 = Qword;       {-- 8 bytes }

  Str2 = String[2];
  Str3 = String[3];
  Str4 = String[4];
  Str7 = String[7];

  ColourType = (BlackPieces, WhitePieces);

  PieceType  = (NoPiece,
                WPawn, WKnight, WBishop, WRook, WQueen, WKing,
                BPawn, BKnight, BBishop, BRook, BQueen, BKing);

{typed}CONST
  PieceValue : Array[PieceType] OF Int16 =         {-- the value of each piece }
               (  0,                               {-- no piece     }
                100, 300, 325, 525, 950, 10000,    {-- white pieces }
                100, 300, 325, 525, 950, 10000);   {-- black pieces }

{typed}CONST  {-- used in debugging }
  PieceTypeString : ARRAY[PieceType] OF Str7 = ('NoPiece',
                                                'WPawn', 'WKnight', 'WBishop', 'WRook', 'WQueen', 'WKing',
                                                'BPawn', 'BKnight', 'BBishop', 'BRook', 'BQueen', 'BKing');


TYPE
  {-- the user commands }
  CmdType = (NoCmd, NewGameCmd, BackCmd,  MoveCmd, PlayCmd,
                    SetUpCmd,   WriteCmd, ReadCmd, PVCmd,
                    TimeCmd,    DebugCmd, EvalCmd, QuitCmd);



{--------------------- The main data structure ------------------------------}
TYPE
  {-- the board representation (0x88)...which piece is on each square of the board }
  SquareType = 0..MaxSquares;
  BoardType  = ARRAY[SquareType] OF PieceType;

  {-- the piece lists....the square a piece is on }
  PieceListType = ARRAY[0..1] OF Byte;     {-- allows for 2 of each piece }
  PawnListType  = ARRAY[0..7] OF Byte;     {-- the 8 pawns }

  {-- the main data structure for each ply }
  DataType = RECORD                        {-- contains all the info for a node   }
               Board      : BoardType;     {-- 0x88 board..what piece is on a square }

               W_King,
               B_King     : SquareType;    {-- what squares the kings occupy }

               W_Pawns,
               B_Pawns    : PawnListType;  {-- list of where each pawn is  }
                                           {-- it will NOT be in file order  }
               W_Queen,
               B_Queen,
               W_Rook,
               B_Rook,
               W_Bishop,
               B_Bishop,
               W_Knight,
               B_Knight       : PieceListType; {-- list of where each piece is  }

               WcanCastShort,                  {-- white/black castled ? }
               WcanCastLong,
               BcanCastShort,
               BcanCastLong   : Boolean;

               WhtPwnHashKey,
               BlaPwnHashKey,
               ZorbistKey     : Word64;        {-- for this board position }

               {-- record when a pawn moves from its home two squares forward. This simplifies }
               {-- generating enpausant captures. Will be either $30..$37, or $40..$47 }
               EnpasauntSq : SquareType;
             END;


VAR
  Data : ARRAY[0..MaxPly] OF DataType;  {--the board representation and piece lists at each ply }
  Ply  : Byte = 0;                      {-- root is 0 ply }



{------------------- The move data structure --------------------------------}
{-- the move is generally kept as its string representation and is converted to
    its full representation when needed }
TYPE
  MoveType = RECORD               {-- Move Definitions....info for each move }
               MoveStr   : Str4;
               MoveValue : Int16; {-- used in move gen sorting }
             END;

  MoveRecType = RECORD            {-- used when converting the move string to full description }
                  FromSq,
                  ToSq          : SquareType;
                  MovingPiece,
                  CapturedPiece : PieceType;
                  Enpasaunt,
                  Castling,
                  Promotion     : Boolean;
                END;

  KillerType    = ARRAY[0..1, 0..MaxPly] OF MoveType;        {-- two killer moves per ply }
  PVtype        = ARRAY[0..MaxPly, 0..MaxPly] OF Str4;       {-- the PV triangular array }
  HistoryType   = ARRAY[PieceType, 0..MaxSquares] OF int16;  {-- the good move history }
  SaveNodesType = ARRAY[0..MaxPly] OF Word32;                {-- for accounting purposes }

VAR
  MoveNmbr  : 1..MaxGameMoves = 1;                           {-- the games move number }
  Moves     : ARRAY[0..MaxGameMoves] OF MoveType;            {-- list of the moves in the game }

  {-- holds the ZorbistKeys for the whole game, used for 50 move rule }
  GameHistoryIndex  : 0..MaxGameMoves = 0;
  GameHistoryKeys   : ARRAY[0..MaxGameMoves] OF Word64; {-- zorbist key for each position }
  MovesSinceCapture : Byte = 0;

  KillerMoves  : KillerType;   {-- list of two killer moves for each ply }
  HistoryTable : HistoryType;  {-- used for move ordering }
  PV           : PVtype;       {-- holds the principle variation }

{--------------------------------Game state variables-------------------------}
VAR
  ComputerPlays : ColourType = BlackPieces;
  Wtm           : Boolean    = True;        {-- white to move }
  HumansMove    : Boolean    = True;
  EndGame       : Boolean    = False;
  Debug         : Boolean    = False;
  Cmd           : CmdType    = NoCmd;

  MoveTime      : int16 =  60;             {-- default is 60 seconds }
  FixedDepth    : Byte  = 255;             {-- 255 when using move time }
  FiftyMoveCnt  : Byte  =   0;             {-- moves without a capture }
  Nodes,
  QNodes        : Word64;



IMPLEMENTATION

BEGIN

END.

