UNIT Search;
{-- the recursive search off the board }

INTERFACE
USES
  Globals;


FUNCTION CalcCompMove : Str4;


IMPLEMENTATION
USES
  Crt,
  Utils,
  Move,
  GenMoves,
  Clocks,
  TT,
  Zorbist,
  Debugging,
  Eval;


{$ASSERTIONS ON}

CONST
  MateScore = 10000;


VAR
  MaxDepth    : int8;
  PVscore     : Int16;

  {-- vars used for accounting }
  Time        : int32;
  BetaCuts,
  NullCuts,
  TThits,
  ExactTThits : Word32;
  PlyTime     : int32;
  SaveNodes   : SaveNodesType;
  TimesUp,
  Abort       : Boolean;



PROCEDURE WriteInfoToScreen;
{-- writes some useful info from the search for debugging and analysing }
BEGIN
  GoToXY(1, 1);
  IF EndGame THEN
    Write('*');

  Window(49, 2, 79, 12);
  ClrScr;

  Writeln('Nodes:' ,  Nodes  : 16);
  Write  ('QNodes:',  QNodes : 15);

  IF Nodes > 0 THEN
    Writeln(Round(100*QNodes/Nodes) : 3, '%');
  IF Time > 0 THEN
    Writeln('NPS:', Nodes DIV Time : 18)
  ELSE
    Writeln('NPS:', 0 : 18);

  Writeln('EBF:',       AverageEBF(SaveNodes, MaxDepth) : 18 : 1);   {-- branching factor }
  Writeln('Beta Cuts:', BetaCuts : 12);
  Writeln('Null Cuts:', NullCuts : 12);
  Writeln('TT Hits:',   ExactTThits : 6, '/', TThits : 7);
  Writeln('Time Used:', Time : 12);
  Writeln('Depth:',     MaxDepth : 16);

  IF (Abs(PVscore) > 9980) THEN   {-- a mate in x moves }
    IF Abs(PVscore) = 9999 THEN
      Write('***** Check Mate *****')
    ELSE
      Write('Mate in:', (10000 - Abs(PVscore)) : 14)
  ELSE
    Write('Score:', -PVscore : 16);

  WritePV(MaxDepth);
END;


PROCEDURE UpDatePV(M : Str4; d : Byte);
VAR
  i : Byte;
BEGIN
  PV[d, d] := M;
  FOR i := 1 TO MaxDepth DO
    PV[d, d+i] := PV[d+1, d+i];
END;


FUNCTION Qsearch(Alpha, Beta : Int16) : Int16;
VAR
  Score    : Int16;
  Move,
  NumMoves : Byte;
  MoveStr  : Str4;

BEGIN
  Score := Evaluation(Alpha, Beta); {-- get the stand pat score }

  IF Ply = MaxPly THEN              {-- cant search for ever }
    Exit(Score);

  IF Score >= Beta THEN             {-- cutoff, we wont be going down this line }
    BEGIN
      Inc(BetaCuts);
      Exit(Beta);
    END;

  IF Score > Alpha THEN
    Alpha := Score;              {-- and update alpha }

  NumMoves := GenCaptureMoves;   {-- only get the capture moves }

  FOR Move := 1 TO NumMoves DO   {-- will not enter the loop if there are no moves }
    BEGIN
      MoveStr := GetAMove(NumMoves);

      {-- copy the data from the current ply, so we can easily takeback the move }
      Inc(Ply);
      Data[Ply] := Data[Ply - 1];

      MakeMove(MoveStr, True);
      Score := -QSearch(-Beta, -Alpha);  {-- the recursive call }
      UnMakeMove;  {-- only needs to decrement the ply to return to the previous board position }

      {-- now tally up the score }
      IF Score > Alpha THEN
        BEGIN
          Alpha := Score;
          IF Alpha >= Beta THEN
            BEGIN
              Inc(BetaCuts);
              Exit(Beta);
            END;
         END;
    END;  {-- of moves loop }

  Qsearch := Alpha;
END;


FUNCTION Search(Alpha, Beta : Int16; Depth : int8; NullAllowed : Boolean) : Int16;
{-- NegaMax always returns the score from the side to move point of view.
    So if white calls it, a positive score is good for white. If black calls it, a positive score
    is good for black }
VAR
  MoveStr     : Str4;
  Move,
  NumMoves    : Byte;
  Check,
  NoMoves,
  FailLowNode,
  CaptureMove : Boolean;
  Score       : Int16;
  TTentry     : TTtype;
  TTmove      : Str4;

BEGIN
  IF Abort THEN
    Exit(0);

  IF Ply = MaxPly THEN   {-- gone too deep, so stop here }
    Exit(Evaluation(Alpha, Beta));

  IF (Ply > 0) AND ThreeFoldDraw THEN  {-- try to avoid this and make a different move }
    Exit(0);

{*--------------------------------Transposition stuff---------------------------}

  TTmove := '';  {-- unless we prove otherwise }

  {-- try to find a transposition node first }
  TTentry.Key := Data[Ply].ZorbistKey;     {-- set the key }

  IF GetFromTT(TTentry) AND ValidMove(TTentry.Move) THEN  {-- then we have visited this node before }
    BEGIN
      Inc(TThits);

      IF (TTentry.Draft >= Depth) THEN     {-- usable info }
        BEGIN
          IF (TTentry.Node = Exact) THEN   {-- perfect match }
            BEGIN
              Inc(ExactTThits);
              UpDatePV(TTentry.Move, Ply);
              {TTmove := TTentry.Move;}
              Exit(TTentry.Score);
            END
          ELSE
            IF (TTentry.Node = FailLow) AND (TTentry.Score <= Alpha) THEN
              Exit(Alpha)
            ELSE
              IF (TTentry.Node = FailHigh) AND (TTentry.Score >= Beta) THEN
                Exit(Beta)
              ELSE
                IF (TTentry.Move = 'NULL') AND (TTentry.Score >= Beta) THEN
                  Exit(Beta);
        END    {-- height >= depth }
      ELSE
        BEGIN  {-- found the node, but does not have the needed depth, we can still try the move early }
          IF TTentry.Node <> FailLow THEN  {-- hoping for an early beta cut }
            TTmove := TTentry.Move;
        END;
    END {-- TT hit }
  ELSE
    IF (PV[Ply, Ply] <> '') AND ValidMove(PV[Ply, Ply]) THEN  {-- TT move may have been overwritten }
      TTmove := PV[Ply, Ply];


  {-- if we are in check we will search a bit deeper }
  Check := InCheck(Wtm);
  IF Check THEN
    Inc(Depth);

  {-- terminal node so get the eval via the q search }
  IF (Depth < 1) THEN
    Exit(QSearch(Alpha, Beta));


  {----- now try a null move looking for a cutoff and an early exit -----}
  IF NullAllowed AND
     NOT Check   AND          {-- you cant pass when in check }
     (Ply > 3)  (* AND
     NOT OnlyPawns(Wtm) *) THEN  {-- risk zugswag....it will be rare }
    BEGIN
      IF Depth > 2 THEN
        BEGIN
          SwapSides;
          Score := -Search(-Beta, -Beta + 1, Depth - 2, False);  {-- shallow null window search centered around Beta }
          SwapSides;

          IF (Score >= Beta) THEN   {-- we beat beta so we wont be going down this line }
            BEGIN
              Inc(BetaCuts);
              Inc(NullCuts);
              AddToTT(Data[Ply].ZorbistKey, Score, FailHigh, Depth, 'NULL');
              Exit(Beta);
            END;
        END;
  END;

{------------------------- now search all other moves -------------------------}

  NoMoves     := True;                 {-- unless proven otherwise }
  FailLowNode := True;                 {-- unless proven otherwise }
  NumMoves    := GenAllMoves(TTmove);  {-- and put TT move first }

  FOR Move := 1 TO NumMoves DO         {-- the move loop: will not enter the loop if there are no moves }
    BEGIN
      MoveStr     := GetAMove(NumMoves);   {-- pick a move to search }
      CaptureMove := NOT (Data[Ply].Board[CoOrdOfToSq(MoveStr)] = NoPiece);

      {-- copy the data to the next ply }
      Inc(Ply);
      Data[Ply] := Data[Ply - 1];

      MakeMove(MoveStr, False);

      IF InCheck(NOT Wtm) THEN         {-- if still in check then this move is not legal }
        BEGIN                          {-- skip all moves that dont get you out of check }
          UnMakeMove;
          Continue;                    {-- and get the next move }
        END
      ELSE
        NoMoves := False;

    {----------------------- now search the move ----------------------------------------}

    IF (move = 1) OR (Depth < 5) OR endgame THEN             {-- PV search...the first move "should" be the best move }
      score := -search(-beta, -alpha, depth-1, False)        {-- dont do null move in PV, or the first 4 ply }
    ELSE
      {-- all others are search in a zero window }
      IF (MoveList[Ply, Move].MoveValue > 198) THEN          {-- LMR stage 1....captures }
        score := -search(-(alpha+1), -alpha, depth-2, True)
      ELSE
        score := -search(-(alpha+1), -alpha, depth-3, True); {-- LMR stage 2....quiet moves via pst+history }

    IF (score > alpha) AND (score < beta) THEN               {-- if we are wrong we must research }
      score := -search(-beta, -alpha, depth-1, True);

    {-------------------------------------------------------------------------------------}

    UnMakeMove; {-- only needs to decrement the ply to return to the previous board position }

    {-- now tally up the score }
	   IF (Score > Alpha) THEN              {-- a better move }
   	  BEGIN
   	    Alpha       := Score;            {-- alpha is best score so far }
        FailLowNode := False;            {-- obviously since score > alpha }

  	     IF NOT CaptureMove THEN
  	       AddToHistory(MoveStr, Depth);  {-- save what is a good move }

  	     IF (Alpha >= Beta) THEN          {-- a beta cut off }
  	       BEGIN
            Inc(BetaCuts);
(*   			      IF NOT CaptureMove THEN
              AddToKillers(Ply, MoveList[Ply, Move]); *)
  			       AddToTT(Data[Ply].ZorbistKey, Score, FailHigh, Depth, MoveStr);
  			       Exit(Beta);
  		      END;

  	     {-- >alpha but <beta is a PV node }
  		    UpDatePV(MoveStr, Ply);
  		    AddToTT(Data[Ply].ZorbistKey, Score, Exact, Depth, MoveStr);
  	   END;  {-- tallying up }

	  {-- this has been an "all node"...all moves have been searched and failed low }
	  IF FailLowNode AND NOT NoMoves THEN
		   AddToTT(Data[Ply].ZorbistKey, Score, FailLow, Depth, '');    {-- there is no good move to store }

	  {-- has the operator pressed the p key }
	  IF KeyPressed AND (ReadKey = 'p') THEN
		   BEGIN
		     Abort := True;
		     Exit(0);
		   END;

	 END;      {-- of the moves loop }

  IF NoMoves THEN
    IF NOT InCheck(Wtm) THEN   {-- stalemate }
      Search := 0
    ELSE
      Search := -MateScore + Ply
  ELSE
    Search := Alpha;
END;


FUNCTION IsTimeUp : Boolean;
{-- true if times up }
VAR
  AveEBF   : Real;
  HalfTime : int32;
BEGIN
  IsTimeUp := True;                     {-- unless its not }
  Time     := Clock(Elapsed);           {-- the time used so far }
  HalfTime := MoveTime DIV 2;           {-- half the allowed time }

  IF Time <= HalfTime THEN
    IsTimeUp := False
  ELSE
    BEGIN
      AveEBF   := AverageEBF(SaveNodes, MaxDepth);

      IsTimeup :=  (Time > HalfTime) OR ( (MoveTime - Time) < (PlyTime * AveEBF) );
                   {-- time taken is > half time allowed OR
                                          Time left < guess of the next ply time }
    END;
END;


PROCEDURE InitSearch;
BEGIN
  GoToXY(49, 13);
  Write('I''m Thinking.......');
  GoToXY(49, 14);
  ClrEol;
  Write('Your Move: ', Moves[MoveNmbr - 1].MoveStr, '     ');

  {-- zero tables }
  FillChar(PV, SizeOf(PV), 0);
  FillChar(KillerMoves, SizeOf(KillerMoves),  0);
  FillChar(HistoryTable, SizeOf(HistoryTable), 0);
  InitTTs;

  {-- and counters }
  Nodes       :=  1;  {-- to avoid divide by zero }
  QNodes      :=  1;
  BetaCuts    :=  0;
  NullCuts    :=  0;
  TThits      :=  0;
  ExactTThits :=  0;
  PVscore     :=  0;

  TimesUp  := False;
  Abort    := False;          {-- abort ply if 'p' pressed }
  EndGame  := GetEndGame;     {-- is it the end game yet ? }

  {-- pawns are worth more in the end game }
  IF EndGame THEN
    BEGIN
      PieceValue[WPawn] := 125;
      PieceValue[BPawn] := 125;

      {-- setup tradedown process when ahead by making the opponents pieces more valuable }
      IF (WhiteMaterial >= BlackMaterial) AND
         ( (NumOfPawns(WhitePieces) - NumOfPawns(BlackPieces) ) > 1) THEN
        AlterPieceValues('B')
      ELSE
        IF (BlackMaterial >= WhiteMaterial) AND
           ( (NumOfPawns(BlackPieces) - NumOfPawns(WhitePieces) ) > 1) THEN
          AlterPieceValues('W');
    END;

  Time     := Clock(Start);
  MaxDepth := 3;               {-- best value....dont change this }
  PlyTime  := 0;

END;


FUNCTION CalcCompMove : Str4;
BEGIN
  InitSearch;

  WHILE NOT TimesUp AND                                 {-- for time per move }
        NOT Abort   AND                                 {-- 'p' key has been pressed }
        (MaxDepth < FixedDepth) AND                     {-- for fixed ply per move }
        (Abs(PVscore) < (MateScore - 20)) DO            {-- forced mate found }
    BEGIN                                               {-- the iterative deepen loop }
      Inc(MaxDepth);                                    {-- search the next ply...iterative deepening }

      PVscore := Search(-30000, 30000, MaxDepth, True); {-- call the search }

      SaveNodes[MaxDepth] := Nodes + Qnodes;            {-- the nodes for each ply searched }

      {-- do the time management }
      Time    := Clock(Elapsed);     {-- get the time }
      PlyTime := Time - PlyTime;     {-- time for the latest ply }
      TimesUp := IsTimeUp;           {-- is it ?....will always complete the current ply }

      IF Abs(PVscore) <> MateScore THEN
        WriteInfoToScreen;
    END;  {-- of iterative deepening }

  {-- check that we haven't stuffed up during the search }
  Assert((Data[Ply].ZorbistKey    = CalcBrdKey),       'error zorbist key');
  Assert((Data[Ply].WhtPwnHashKey = CalcWhtPwnBrdKey), 'error white pawn hash key');
  Assert((Data[Ply].BlaPwnHashKey = CalcBlaPwnBrdKey), 'error black pawn hash key');

  IF Abs(PVscore) = MateScore THEN
    CalcCompMove := 'MATE'
  ELSE
    CalcCompMove := PV[0, 0];   {-- first move at the top of the PV }

{  WriteHistoryTable;}
(*  WalkThePV(MaxDepth - 1);*)
END;



BEGIN
END.
